/*

Greeter widget for kdm

Copyright (C) 1997, 1998, 2000 Steffen Hansen <hansen@kde.org>
Copyright (C) 2000-2004 Oswald Buddenhagen <ossi@kde.org>
Copyright (C) 2011 ROSA Lab. <ural.mullabaev@rosalab.ru>


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "kgreeter.h"
#include "kconsole.h"
#include "kdmconfig.h"
#include "kdmclock.h"
#include "kdm_greet.h"
#include "kdmshutdown.h"
#include "themer/kdmthemer.h"
#include "themer/kdmitem.h"
#include "themer/kdmlabel.h"

#include <KColorScheme>
#include <KConfigGroup>
#include <klocale.h>
#include <kseparator.h>
#include <kstandarddirs.h>
#include <kstringhandler.h>

#include <QAction>
#include <QBuffer>
#include <QDir>
#include <QFile>
#include <QImageReader>
#include <QKeyEvent>
#include <QLabel>
#include <QListWidget>
#include <QListWidgetItem>
#include <QMenu>
#include <QMovie>
#include <QPainter>
#include <QLineEdit>
#include <QPushButton>
#include <QShortcut>
#include <QStyle>
#include <QTextLayout>
#include <QParallelAnimationGroup>
#include <QX11Info>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <X11/Xlib.h>
#include <fixx11h.h>

#define LIST_IMAGE_SIZE 48
#define SLIDER_IMAGE_SIZE 150
#define SELECTED_IMAGE_RATIO 1.5

UserListView::UserListView(QWidget *parent)
    : QListWidget(parent)
    , cachedSizeHint(-1, 0)
{
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Ignored);
    setUniformItemSizes(true);
    setVerticalScrollMode(ScrollPerPixel);
    setIconSize(QSize(LIST_IMAGE_SIZE, LIST_IMAGE_SIZE));
    setAlternatingRowColors(true);
}

QSize
UserListView::sizeHint() const
{
    if (!cachedSizeHint.isValid()) {
        ensurePolished();
        QStyleOptionViewItem vo(viewOptions());
        QAbstractListModel *md(static_cast<QAbstractListModel *>(model()));
        uint maxw = 0, h = 0;
        for (int i = 0, rc = md->rowCount(); i < rc; i++) {
            QSize sh = itemDelegate()->sizeHint(vo, md->index(i));
            uint thisw = sh.width();
            if (thisw > maxw)
                maxw = thisw;
            h += sh.height();
        }
        cachedSizeHint.setWidth(
            style()->pixelMetric(QStyle::PM_ScrollBarExtent) +
            (frameWidth() +
             (style()->styleHint(QStyle::SH_ScrollView_FrameOnlyAroundContents) ?
              style()->pixelMetric(QStyle::PM_DefaultFrameWidth) : 0)) * 2 +
            maxw);
        cachedSizeHint.setHeight(frameWidth() * 2 + h);
    }
    return cachedSizeHint;
}

void
UserListView::keyPressEvent(QKeyEvent *event)
{
    switch (event->key()) {
    case Qt::Key_Enter:
    case Qt::Key_Return:
        if (currentItem())
            emit itemDoubleClicked(currentItem());
        event->accept();
        break;
    case Qt::Key_Space:
        if (currentItem())
            emit itemClicked(currentItem());
        event->accept();
        break;
    default:
        QListWidget::keyPressEvent(event);
        break;
    }
}

void
UserListView::mousePressEvent(QMouseEvent *event)
{
    m_suppressClick = false;
    QListWidget::mousePressEvent(event);
}

void
UserListView::mouseReleaseEvent(QMouseEvent *event)
{
    if (m_suppressClick)
        m_suppressClick = false;
    else
        QListWidget::mouseReleaseEvent(event);
}

void
UserListView::mouseDoubleClickEvent(QMouseEvent *event)
{
    m_suppressClick = true;
    QListWidget::mouseDoubleClickEvent(event);
}


UserListViewItem::UserListViewItem(UserListView *parent, const QString &text,
                    		   const QIcon &icon, const QString &username)
    : QListWidgetItem(parent)
    , login(username)
{
    setIcon(icon);
    setText(text);
    parent->cachedSizeHint.setWidth(-1);
}


UserSliderView::UserSliderView(QWidget *parent)
    : UserListView(parent)
{
    setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Minimum);
    setIconSize(QSize(SLIDER_IMAGE_SIZE * SELECTED_IMAGE_RATIO, SLIDER_IMAGE_SIZE * SELECTED_IMAGE_RATIO));
    setUniformItemSizes(false);
    setAlternatingRowColors(false);
    setViewMode(QListView::IconMode);
    setSelectionMode(QAbstractItemView::NoSelection);
    setWrapping(false);
    setSpacing(10);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setAcceptDrops(false);
}

QWidget*
UserSliderView::itemEditor(const int row)
{
    return itemEditors.value(row);
}

void
UserSliderView::setItemEditor(const int row, QWidget *widget)
{
    itemEditors.resize(count());
    itemEditors.replace(row, widget);
}

void
UserSliderView::layoutItems()
{
    QStyleOptionViewItem vo(viewOptions());
    QAbstractListModel *md(static_cast<QAbstractListModel *>(model()));
    int rc = model()->rowCount();
    int w = 0;
    for (int i = 0; i < rc; i++) {
        QSize sh = itemDelegate()->sizeHint(vo, md->index(i));
        w += sh.width();
    }
    int lw = size().width();
    int lh = size().height();
    int s = spacing();
    int os = (lw - w - s * (rc-1)) / 2;
    int l = os;

    for (int i = 0; i < rc; i++) {
        QSize sh = itemDelegate()->sizeHint(vo, md->index(i));
	setPositionForIndex(QPoint(l, (lh - sh.height()) / 2), md->index(i));
	l += sh.width() + spacing();
    }
    updateGeometries();
}

void
UserSliderView::updateGeometries()
{
    QListWidget::updateGeometries();
}

void
UserSliderView::resizeEvent(QResizeEvent *event)
{
    QListWidget::resizeEvent(event);

    const int margin  = 3; //style()->pixelMetric(QStyle::PM_FocusFrameHMargin) + 1;
    const int text    = 24;
    QSize size = QSize(SLIDER_IMAGE_SIZE + 16 * margin,
		       SLIDER_IMAGE_SIZE + 18 * margin + text);
    for (int i = 0; i < model()->rowCount(); i++) {
	item(i)->setSizeHint(size);
    }
    layoutItems();
}

void
UserSliderView::dragEnterEvent (QDragEnterEvent *event)
{
    event->ignore();
}

void
UserSliderView::dragMoveEvent (QDragMoveEvent *event)
{
    event->ignore();
}

void
UserSliderView::dragLeaveEvent (QDragLeaveEvent *event)
{
    event->ignore();
}

void
UserSliderView::dropEvent (QDropEvent *event)
{
    event->ignore();
}

UserSliderViewItem::UserSliderViewItem(UserSliderView *parent, const QString &text,
                 const QIcon &icon, const QString &username)
    : QObject()
    , UserListViewItem(parent, text, icon, username)
{
    parent->openPersistentEditor(this);
}

UserSliderViewItem::~UserSliderViewItem()
{
}


UserSliderViewItemAnimation::UserSliderViewItemAnimation(QObject *parent)
    : QPropertyAnimation(parent)
{
}

void
UserSliderViewItemAnimation::updateCurrentValue(const QVariant &value)
{
    QPropertyAnimation::updateCurrentValue(value);
    emit sizeChanged();
}

void
UserItemDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option,
                         const QModelIndex &index) const
{
    QStyleOptionViewItemV4 opt = option;
    initStyleOption(&opt, index);

    const QWidget *widget = opt.widget;
    QStyle *style = widget->style();

    if (const QStyleOptionViewItemV4 *vopt = qstyleoption_cast<const QStyleOptionViewItemV4 *>(&opt)) {
        painter->save();
        painter->setClipRect(opt.rect);
	painter->setRenderHint(QPainter::Antialiasing);
	const int margin = 3; //widget->style()->pixelMetric(QStyle::PM_FocusFrameHMargin, 0, widget) + 1;
	const int offset = 2 * margin;
	const int shadow = 4 * margin;
	const int frame1  = 2 * margin;
	const int frame2  = 2;
	const int radius = 2 * margin;
	const int spacing  = 2 * margin;
	const int text  = 24; /*vopt->fontMetrics.height()*/

        QRect iconRect = QRect(vopt->rect.left() + offset + shadow + frame1 + frame2,
        		       vopt->rect.top() + offset + shadow + frame1 + frame2,
        		       vopt->rect.width() - 2 * (offset + shadow + frame1 + frame2),
        		       vopt->rect.width() - 2 * (offset + shadow + frame1 + frame2));
        QRect textRect = QRect(vopt->rect.left(),
        		       vopt->rect.top() + vopt->rect.height() - text - spacing - offset,
        		       vopt->rect.width(),
        		       text);

        QTextLayout textLayout;
        textLayout.setFont(vopt->font);
        textLayout.setText(vopt->text);
        textLayout.beginLayout();
        QTextLine line = textLayout.createLine();
        line.setLineWidth(textRect.width());
        textLayout.endLayout();
        textRect.setWidth(line.naturalTextWidth() + 4 * margin);
        textRect.moveLeft(vopt->rect.left() + vopt->rect.width() / 2 - textRect.width() / 2);

        // draw the background
        QPalette::ColorGroup cg = (widget ? widget->isEnabled() : (vopt->state & QStyle::State_Enabled))
                                      ? QPalette::Normal : QPalette::Disabled;
        /*
        if (vopt->showDecorationSelected && (vopt->state & QStyle::State_Selected)) {
            painter->fillRect(vopt->rect, vopt->palette.brush(cg, QPalette::Highlight));
        } else {
            if (vopt->backgroundBrush.style() != Qt::NoBrush) {
        	qDebug("draw the background   != Qt::NoBrush");
                QPointF oldBO = painter->brushOrigin();
                painter->setBrushOrigin(vopt->rect.topLeft());
                painter->fillRect(vopt->rect, vopt->backgroundBrush);
                painter->setBrushOrigin(oldBO);
            }
        */
        QRect shadowLeftTopRect = QRect(vopt->rect.left() + offset,
        			 vopt->rect.top() + offset,
        			 2 * (shadow + radius),
        			 2 * (shadow + radius));
        QRect shadowRightTopRect = QRect(vopt->rect.left() + vopt->rect.width() - offset - 2 * (shadow + radius),
        			 vopt->rect.top() + offset,
        			 2 * (shadow + radius),
        			 2 * (shadow + radius));
        QRect shadowRightBottomRect = QRect(vopt->rect.left() + vopt->rect.width() - offset - 2 * (shadow + radius),
        			 vopt->rect.top() + vopt->rect.height() - 2 * (shadow + radius) - spacing - text - offset,
        			 2 * (shadow + radius),
        			 2 * (shadow + radius));
        QRect shadowLeftBottomRect = QRect(vopt->rect.left() + offset,
        			 vopt->rect.top() + vopt->rect.height() - 2 * (shadow + radius) - spacing - text - offset,
        			 2 * (shadow + radius),
        			 2 * (shadow + radius));

        QRect shadowLeftRect = QRect(vopt->rect.left() + offset,
        			 vopt->rect.top() + offset + shadow + radius,
        			 shadow + radius,
        			 vopt->rect.height() - 2 * (offset + shadow + radius) - spacing - text);
        QRect shadowTopRect = QRect(vopt->rect.left() + offset + shadow + radius,
        			 vopt->rect.top() + offset,
        			 vopt->rect.width() - 2 * (offset + shadow + radius),
        			 shadow + radius);
        QRect shadowRightRect = QRect(vopt->rect.left() + vopt->rect.width() - offset - shadow - radius,
        			 vopt->rect.top() + offset + shadow + radius,
        			 shadow + radius,
        			 vopt->rect.height() - 2 * (offset + shadow + radius) - spacing - text);
        QRect shadowBottomRect = QRect(vopt->rect.left() + offset + shadow + radius,
        			 vopt->rect.top() + vopt->rect.height() - shadow - radius - spacing - text - offset,
        			 vopt->rect.width() - 2 * (offset + shadow + radius),
        			 shadow + radius);

	painter->setPen(QPen(QColor(0, 0, 0, 0), 0));

        QRadialGradient radialGrad;
        radialGrad.setColorAt(0, QColor(0, 0, 0, 48));
        radialGrad.setColorAt(1, QColor(0, 0, 0, 0));
        radialGrad.setRadius(shadow + radius);

        radialGrad.setCenter(shadowLeftTopRect.left() + shadow + radius + 1, shadowLeftTopRect.top() + shadow + radius + 1);
        radialGrad.setFocalPoint(shadowLeftTopRect.left() + shadow + radius + 1, shadowLeftTopRect.top() + shadow + radius + 1);
	painter->setBrush(radialGrad);
        painter->drawPie(shadowLeftTopRect, 90 * 16, 90 * 16);

        radialGrad.setCenter(shadowRightTopRect.right() - shadow - radius - 1, shadowRightTopRect.top() + shadow + radius + 1);
        radialGrad.setFocalPoint(shadowRightTopRect.right() - shadow - radius - 1, shadowRightTopRect.top() + shadow + radius + 1);
	painter->setBrush(radialGrad);
        painter->drawPie(shadowRightTopRect, 0 * 16, 90 * 16);

        radialGrad.setCenter(shadowRightBottomRect.right() - shadow - radius - 1, shadowRightBottomRect.bottom() - shadow - radius - 1);
        radialGrad.setFocalPoint(shadowRightBottomRect.right() - shadow - radius - 1, shadowRightBottomRect.bottom() - shadow - radius - 1);
	painter->setBrush(radialGrad);
        painter->drawPie(shadowRightBottomRect, 270 * 16, 90 * 16);

        radialGrad.setCenter(shadowLeftBottomRect.left() + shadow + radius + 1, shadowLeftBottomRect.bottom() - shadow - radius - 1);
        radialGrad.setFocalPoint(shadowLeftBottomRect.left() + shadow + radius + 1, shadowLeftBottomRect.bottom() - shadow - radius - 1);
	painter->setBrush(radialGrad);
        painter->drawPie(shadowLeftBottomRect, 180 * 16, 90 * 16);

        QLinearGradient linearGrad;
        linearGrad.setColorAt(0, QColor(0, 0, 0, 48));
        linearGrad.setColorAt(1, QColor(0, 0, 0, 0));

        linearGrad.setStart(shadowLeftRect.right(), 0);
        linearGrad.setFinalStop(shadowLeftRect.left(), 0);
	painter->setBrush(linearGrad);
        painter->drawRect(shadowLeftRect);

        linearGrad.setStart(0, shadowTopRect.bottom());
        linearGrad.setFinalStop(0, shadowTopRect.top());
	painter->setBrush(linearGrad);
        painter->drawRect(shadowTopRect);

        linearGrad.setStart(shadowRightRect.left(), 0);
        linearGrad.setFinalStop(shadowRightRect.right(), 0);
	painter->setBrush(linearGrad);
        painter->drawRect(shadowRightRect);

        linearGrad.setStart(0, shadowBottomRect.top());
        linearGrad.setFinalStop(0, shadowBottomRect.bottom());
	painter->setBrush(linearGrad);
        painter->drawRect(shadowBottomRect);

        QRect frame0Rect = QRect(vopt->rect.left() + offset + shadow + frame1,
        			vopt->rect.top() + vopt->rect.width() - (offset + shadow + frame1),
        			vopt->rect.width() - 2 * (offset + shadow + frame1), 
            			vopt->rect.height() - vopt->rect.width() - spacing - text);
	painter->setBrush(QColor(0, 0, 0, 48));
        painter->drawRect(frame0Rect);
        QRect frame1Rect = QRect(vopt->rect.left() + offset + shadow,
        			vopt->rect.top() + offset + shadow,
        			vopt->rect.width() - 2 * (offset + shadow), 
            			vopt->rect.height() - 2 * (offset + shadow) - spacing - text);
	painter->setBrush(QColor(0, 0, 0, 148));
        painter->drawRoundedRect(frame1Rect, radius, radius);
        QRect frame2Rect = QRect(vopt->rect.left() + offset + shadow + frame1,
        			vopt->rect.top() + offset + shadow + frame1,
        			vopt->rect.width() - 2 * (offset + shadow + frame1),
            			vopt->rect.width() - 2 * (offset + shadow + frame1));
	painter->setBrush(QColor(255, 255, 255, 255));
        painter->drawRoundedRect(frame2Rect, margin, margin);

        // draw the icon
        QIcon::Mode mode = QIcon::Normal;
        if (!(vopt->state & QStyle::State_Enabled))
            mode = QIcon::Disabled;
        else if (vopt->state & QStyle::State_Selected)
            mode = QIcon::Selected;
        QIcon::State state = vopt->state & QStyle::State_Open ? QIcon::On : QIcon::Off;
        vopt->icon.paint(painter, iconRect, vopt->decorationAlignment, mode, state);

        // draw the text
        if (!vopt->text.isEmpty()) {
	    painter->setPen(QColor(255, 255, 255, 255));
            drawText(painter, vopt, textRect);
        }
	painter->restore();
    }

}

QWidget *
UserItemDelegate::createEditor(QWidget *parent,
                                    const QStyleOptionViewItem &option,
                                    const QModelIndex &index) const

{
    PasswordEditor *editor = new PasswordEditor(parent);
    connect(editor, SIGNAL(textEntered(const QString &)),
            this, SLOT(slotTextEntered(const QString &)));
    connect(editor, SIGNAL(tabPressed(const int)),
            this, SLOT(slotTabPressed(const int)));

    QStyleOptionViewItemV4 opt = option;
    QWidget *widget = const_cast<QWidget *>(opt.widget);
    UserSliderView *view = static_cast<UserSliderView *>(widget);
    view->setItemEditor(index.row(), editor);
    view->setFocusProxy(editor);

    return editor;
}

void
UserItemDelegate::setEditorData(QWidget * /*editor*/,
                                 const QModelIndex &/*index*/) const
{
}

void
UserItemDelegate::setModelData(QWidget * /*editor*/, QAbstractItemModel * /*model*/,
                                const QModelIndex &/*index*/) const
{
}

void
UserItemDelegate::updateEditorGeometry(QWidget * editor, const QStyleOptionViewItem & option,
			      const QModelIndex & /*index*/) const
{
    QStyleOptionViewItemV4 opt = option;
    const int margin = 3; //opt.widget->style()->pixelMetric(QStyle::PM_FocusFrameHMargin, 0, opt.widget) + 1;
    const int text = 24; //opt.widget->style()->pixelMetric(QStyle::PM_FocusFrameHMargin, 0, opt.widget) + 1;

    int x = option.rect.left() + 8 * margin;
    int y = option.rect.top() + option.rect.width() - 6 * margin;
    int w = option.rect.width() - 16 * margin;
    int h = option.rect.height() - option.rect.width() - 4 * margin - text;
    //qDebug("updateEditorGeometry height %d, rh %d, rw", h, option.rect.height(), option.rect.width());
    editor->setGeometry(QRect(x, y, w, h));
}

void
UserItemDelegate::slotTextEntered(const QString &text)
{
    emit passwordEntered(text);
}

void
UserItemDelegate::slotTabPressed(const int key)
{
    emit tabPressed(key);
}

void
UserItemDelegate::drawText(QPainter *p, const QStyleOptionViewItemV4 *option, const QRect &rect) const
{
    const QWidget *widget = option->widget;
    const int textMargin = 3; //widget->style()->pixelMetric(QStyle::PM_FocusFrameHMargin, 0, widget) + 1;

    QRect textRect = rect.adjusted(textMargin, 0, -textMargin, 0);
    const bool wrapText = option->features & QStyleOptionViewItemV2::WrapText;
    QTextOption textOption;
    textOption.setWrapMode(wrapText ? QTextOption::WordWrap : QTextOption::ManualWrap);
    textOption.setTextDirection(option->direction);
    textOption.setAlignment(QStyle::visualAlignment(option->direction, option->displayAlignment));
    QTextLayout textLayout;
    textLayout.setTextOption(textOption);
    textLayout.setFont(option->font);
    textLayout.setText(option->text);

    qreal heightUsed = 0;
    qreal widthUsed = 0;
    textLayout.beginLayout();
    while (true) {
        QTextLine line = textLayout.createLine();
        if (!line.isValid())
            break;
        line.setLineWidth(textRect.width());
        line.setPosition(QPointF(0, heightUsed));
        heightUsed += line.height();
        widthUsed = qMax(widthUsed, line.naturalTextWidth());
    }
    textLayout.endLayout();
    QSizeF textLayoutSize = QSizeF(widthUsed, heightUsed);

    QString elidedText;
    qreal height = 0;
    qreal width = 0;
    int elidedIndex = -1;
    const int lineCount = textLayout.lineCount();
    for (int j = 0; j < lineCount; ++j) {
        const QTextLine line = textLayout.lineAt(j);
        if (j + 1 <= lineCount - 1) {
            const QTextLine nextLine = textLayout.lineAt(j + 1);
            if ((nextLine.y() + nextLine.height()) > textRect.height()) {
                int start = line.textStart();
                int length = line.textLength() + nextLine.textLength();
                //const QStackTextEngine engine(textLayout.text().mid(start, length), option->font);
                elidedText = /*engine*/option->fontMetrics.elidedText(textLayout.text().mid(start, length), option->textElideMode, textRect.width());
                height += line.height();
                width = textRect.width();
                elidedIndex = j;
                break;
            }
        }
        if (line.naturalTextWidth() > textRect.width()) {
            int start = line.textStart();
            int length = line.textLength();
            //const QStackTextEngine engine(textLayout.text().mid(start, length), option->font);
            elidedText = /*engine*/option->fontMetrics.elidedText(textLayout.text().mid(start, length), option->textElideMode, textRect.width());
            height += line.height();
            width = textRect.width();
            elidedIndex = j;
            break;
        }
        width = qMax<qreal>(width, line.width());
        height += line.height();
    }

    const QRect layoutRect = QStyle::alignedRect(option->direction, option->displayAlignment,
                                                 QSize(int(width), int(height)), textRect);
    const QPointF position = layoutRect.topLeft();
    for (int i = 0; i < lineCount; ++i) {
        const QTextLine line = textLayout.lineAt(i);
        if (i == elidedIndex) {
            qreal x = position.x() + line.x();
            qreal y = position.y() + line.y() + line.ascent();
            p->save();
            p->setFont(option->font);
            p->drawText(QPointF(x, y), elidedText);
            p->restore();
            break;
        }
        line.draw(p, position);
    }
}

LineEdit::LineEdit(QWidget *parent)
    : QLineEdit(parent)
{
    setEchoMode(QLineEdit::Password);
    setLangIndicator("us");
    setCapsIndicator(false);
}

void 
LineEdit::setLangIndicator(const QString &lang)
{
    language = lang;
}

void 
LineEdit::setCapsIndicator(const bool caps)
{
    capsLocked = caps;
}

bool 
LineEdit::event(QEvent *event)
{
    if (event->type() == QEvent::KeyPress) {
	QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
	if (keyEvent->key() == Qt::Key_Tab ||
	    keyEvent->key() == Qt::Key_Backtab) {
	    emit tabPressed(keyEvent->key());
	    return true;
	}
    }
    return QLineEdit::event(event);
}

void 
LineEdit::paintEvent(QPaintEvent *event)
{
    QLineEdit::paintEvent(event);

    const int margin  = 3; //style()->pixelMetric(QStyle::PM_FocusFrameHMargin) + 1;
    const int frame   = 20;

    QPainter p(this);
    QRect r = rect();
    QPalette pal = palette();
    QFont fnt = font();
    fnt.setPixelSize((r.height() > 0) ? r.height() * 3 / 8 : 1);
//    fnt.setBold(true);
    QFontMetrics fm(fnt);
    const bool localLang = !language.contains("us");

    QString langText, capsText;
    QRect langFrameRect, langTextRect, capsFrameRect, capsTextRect;

    if (capsLocked) {
	capsText = capsLocked ? QString("A") : QString("a");
        int capsTextWidth = fm.width(capsText);
	int capsTextHeight = fm.height();
    	capsFrameRect = QRect(r.right() - frame - margin - 1, r.top() + margin + 2,
    				    frame, r.height() - 2 * (margin + 2));
        capsTextRect  = QRect(capsFrameRect.left() + (capsFrameRect.width() - capsTextWidth) / 2,
    				    capsFrameRect.top() + (capsFrameRect.height() - capsTextHeight) / 2,
				    capsTextWidth, capsTextHeight);
    }
    if (localLang) {
	langText = language;
	int langTextWidth = fm.width(langText);
	int langTextHeight = fm.height();
	if (capsLocked) {
	    langFrameRect = QRect(r.right() - 2 * (frame + margin) - 1, r.top() + margin + 2,
					frame, r.height() - 2 * (margin + 2));
	    setTextMargins(0, 0, 2 * (frame + margin) + 1, 2);
	}
	else {
	    langFrameRect = QRect(r.right() - frame - margin - 1, r.top() + margin + 2,
					frame, r.height() - 2 * (margin + 2));
	    setTextMargins(0, 0, frame + margin + 1, 2);
	}
	langTextRect  = QRect(langFrameRect.left() + (langFrameRect.width() - langTextWidth) / 2,
				    langFrameRect.top() + (langFrameRect.height() - langTextHeight) / 2,
				    langTextWidth, langTextHeight);
    }
    else {
	if (capsLocked)
	    setTextMargins(0, 0, frame + margin + 1, 2);
	else
	    setTextMargins(0, 0, 0, 2);
    }

    p.save();
    p.setPen(QColor(255, 255, 255, 255));
    p.setFont(fnt);
    if (localLang) {
	p.fillRect(langFrameRect, pal.highlight());
	p.drawText(langTextRect, langText);
    }
    if (capsLocked) {
	p.fillRect(capsFrameRect, pal.highlight());
	p.drawText(capsTextRect, capsText);
    }
    p.restore();


}

PasswordEditor::PasswordEditor(QWidget *parent)
    : QWidget(parent)
{
    layout = new QHBoxLayout(this);
    layout->setMargin(0);
    edit = new LineEdit(this);
    edit->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    connect(edit, SIGNAL(returnPressed()), this, SLOT(slotReturnPressed()));
    connect(edit, SIGNAL(tabPressed(const int)), this, SLOT(slotTabPressed(const int)));
    layout->addWidget(edit);
    button = new QPushButton(">", this);
    button->setFocusPolicy(Qt::NoFocus);
    button->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    button->setMaximumHeight(edit->height() - 1);
    button->setMaximumWidth(edit->height() - 1);
    connect(button, SIGNAL(clicked()), this, SLOT(slotReturnPressed()));
    layout->addWidget(button);
    setLayout(layout);
    setFocusProxy(edit);
}

PasswordEditor::~PasswordEditor()
{
    delete button;
    delete edit;
    delete layout;
}

LineEdit* 
PasswordEditor::lineEdit() const
{
    return edit;
}

void 
PasswordEditor::slotReturnPressed()
{
    emit textEntered(edit->text());
}

void 
PasswordEditor::slotTabPressed(const int key)
{
    emit tabPressed(key);
}

int KGreeter::curPlugin = -1;
PluginList KGreeter::pluginList;

KGreeter::KGreeter(bool framed)
    : inherited(framed)

    , imageSize(LIST_IMAGE_SIZE)
    , dName(dname)
    , userView(0)
    , userList(0)
    , nNormals(0)
    , nSpecials(0)
    , curPrev(0)
    , prevValid(true)
    , needLoad(false)
{
    stsGroup = new KConfigGroup(KSharedConfig::openConfig(_stsFile),
                                "PrevUser");

    if (_userList) {
        userView = new UserListView(this);
        connect(userView, SIGNAL(itemClicked(QListWidgetItem *)),
                SLOT(slotUserClicked(QListWidgetItem *)));
        connect(userView, SIGNAL(itemDoubleClicked(QListWidgetItem *)),
                SLOT(accept()));
    }
    if (_userCompletion)
        userList = new QStringList;
    if (userView || userList)
        insertUsers();

    sessMenu = new QMenu(this);
    connect(sessMenu, SIGNAL(triggered(QAction *)),
            SLOT(slotSessionSelected()));

    sessGroup = new QActionGroup(this);
    insertSessions();

    if (curPlugin < 0) {
        curPlugin = 0;
        pluginList = KGVerify::init(_pluginsLogin);
    }
}

KGreeter::~KGreeter()
{
    hide();
    delete userList;
    delete verify;
    delete stsGroup;
}

#define FILE_LIMIT_ICON 200 	//U 20
#define FILE_LIMIT_IMAGE 200
#define PIXEL_LIMIT_ICON 300	//U 100
#define PIXEL_LIMIT_IMAGE 300

// replace this with a simple !access(..., X_OK) once we run with non-root real uid
static bool
dirAccessible(const char *dir)
{
    struct stat st;

    if (stat(dir, &st))
        return false;
    return (st.st_mode & S_IXOTH) != 0;
}

static bool
loadFace(QByteArray &fn, QIcon &icon, const int size, const QByteArray &pp, bool complain = false)
{
    int fd, ico;
    if ((fd = open(fn.data(), O_RDONLY | O_NONBLOCK)) < 0) {
        if (errno != ENOENT) {
            if (pp.isEmpty() || dirAccessible(pp.data()))
                (complain ? logError : logInfo)
                    ("Cannot load %s: %m\n", fn.data());
            return false;
        }
        fn.chop(5);
        if ((fd = open(fn.data(), O_RDONLY | O_NONBLOCK)) < 0) {
        if ((complain || errno != ENOENT) &&
                (pp.isEmpty() || dirAccessible(pp.data())))
                (complain ? logError : logInfo)
                    ("Cannot load %s: %m\n", fn.data());
            return false;
        }
        ico = 0;
    } else {
        ico = 1;
    }
    QFile f;
    f.open(fd, QFile::ReadOnly);
    int fs = f.size();
    if (fs > (ico ? FILE_LIMIT_ICON : FILE_LIMIT_IMAGE) * 1000) {
        logWarn("%s exceeds file size limit (%dkB)\n",
                fn.data(), ico ? FILE_LIMIT_ICON : FILE_LIMIT_IMAGE);
        return false;
    }
    QByteArray fc = f.read(fs);
    ::close(fd);
    int is = size * SELECTED_IMAGE_RATIO;
    QBuffer buf(&fc);
    buf.open(QBuffer::ReadOnly);
    QImageReader ir(&buf);
    QSize sz = ir.size();
    int lim = ico ? PIXEL_LIMIT_ICON : PIXEL_LIMIT_IMAGE;
    if (sz.width() > lim || sz.height() > lim) {
        logWarn("%s exceeds image dimension limit (%dx%d)\n",
                fn.data(), lim, lim);
        return false;
    }
    sz.scale(is, is, Qt::KeepAspectRatio);
    ir.setScaledSize(sz);
    QImage p = ir.read();
    if (p.isNull()) {
        logInfo("%s is no valid image\n", fn.data());
        return false;
    }
    if (p.width() < is) {
        QImage np(is, p.height(), QImage::Format_ARGB32);
        np.fill(0);
        QPainter pnt(&np);
        pnt.drawImage((is - p.width()) / 2, 0, p);
        p = np;
    }
    icon.addPixmap(QPixmap::fromImage(p), QIcon::Normal);
    return true;
}

static void
getImage(const QIcon &default_icon, QIcon &i, const int size, struct passwd *ps)
{
    int dp = 0, nd = 0;
    if (_faceSource == FACE_USER_ONLY ||
        _faceSource == FACE_PREFER_USER)
        dp = 1;
    if (_faceSource != FACE_USER_ONLY &&
        _faceSource != FACE_ADMIN_ONLY)
        nd = 1;
    do {
        dp ^= 1;
        QByteArray pp, fn;
        if (!dp) {
            fn = pp = QByteArray(ps->pw_dir);
            fn += '/';
            fn += ".face.icon";
        } else {
            fn = QFile::encodeName(_faceDir);
            fn += '/';
            fn += ps->pw_name;
    	    fn += ".png";
        }
        if (loadFace(fn, i, size, pp))
            return;
    } while (--nd >= 0);
    i = default_icon;
    return;
}

void
KGreeter::insertUser(const QIcon &default_icon,
                     const QString &username, struct passwd *ps)
{
    if (userList)
        userList->append(username);
    if (!userView)
        return;

    QIcon icon;
    getImage(default_icon, icon, imageSize, ps);

    QString realname = KStringHandler::from8Bit(ps->pw_gecos);
    realname.truncate(realname.indexOf(',') & (~0U >> 1));
    if (realname.isEmpty() || realname == username) {
        new UserListViewItem(userView, username, icon, username);
    } else {
        realname.append("\n").append(username);
        new UserListViewItem(userView, realname, icon, username);
    }
}

class UserList {
  public:
    UserList(char **in);
    bool hasUser(const char *str) const { return users.contains(str); }
    bool hasGroup(gid_t gid) const { return groups.contains(gid); }
    bool hasGroups() const { return !groups.isEmpty(); }

  private:
    QSet<gid_t> groups;
    QSet<QByteArray> users;
};

UserList::UserList(char **in)
{
    struct group *grp;

    for (; *in; in++)
        if (**in == '@') {
            if ((grp = getgrnam(*in + 1))) {
                for (; *grp->gr_mem; grp->gr_mem++)
                    users.insert(*grp->gr_mem);
                groups.insert(grp->gr_gid);
            }
        } else {
            users.insert(*in);
        }
}

void
KGreeter::insertUsers()
{
    struct passwd *ps;

    if (!getuid()) {
        if (!(ps = getpwnam("nobody")))
            return;
        if (setegid(ps->pw_gid))
            return;
        if (seteuid(ps->pw_uid)) {
            setegid(0);
            return;
        }
    }

    QIcon default_icon;
    if (userView) {
        QByteArray fn = QFile::encodeName(_faceDir) + "/default.png"/*"/.default.face.icon"*/;
        if (!loadFace(fn, default_icon, imageSize, QByteArray(), true)) {
    	    QImage img;
            img = QImage(imageSize, imageSize, QImage::Format_ARGB32);
            img.fill(0);
            default_icon.addPixmap(QPixmap::fromImage(img), QIcon::Normal);
            img = QImage(imageSize * SELECTED_IMAGE_RATIO, imageSize * SELECTED_IMAGE_RATIO, 
        		QImage::Format_ARGB32);
            img.fill(0);
            default_icon.addPixmap(QPixmap::fromImage(img), QIcon::Selected);
        }
    }
    if (_showUsers == SHOW_ALL) {
        UserList noUsers(_noUsers);
        QSet<QString> dupes;
        for (setpwent(); (ps = getpwent()) != 0;) {
            if (*ps->pw_dir && *ps->pw_shell &&
                (ps->pw_uid >= (unsigned)_lowUserId ||
                 (!ps->pw_uid && _showRoot)) &&
                ps->pw_uid <= (unsigned)_highUserId &&
                !noUsers.hasUser(ps->pw_name) &&
                !noUsers.hasGroup(ps->pw_gid))
            {
                QString username(QFile::decodeName(ps->pw_name));
                if (!dupes.contains(username)) {
                    dupes.insert(username);
                    insertUser(default_icon, username, ps);
                }
            }
        }
    } else {
        UserList users(_users);
        if (users.hasGroups()) {
            QSet<QString> dupes;
            for (setpwent(); (ps = getpwent()) != 0;) {
                if (*ps->pw_dir && *ps->pw_shell &&
                    (ps->pw_uid >= (unsigned)_lowUserId ||
                     (!ps->pw_uid && _showRoot)) &&
                    ps->pw_uid <= (unsigned)_highUserId &&
                    (users.hasUser(ps->pw_name) ||
                     users.hasGroup(ps->pw_gid)))
                {
                    QString username(QFile::decodeName(ps->pw_name));
                    if (!dupes.contains(username)) {
                        dupes.insert(username);
                        insertUser(default_icon, username, ps);
                    }
                }
            }
        } else {
            for (int i = 0; _users[i]; i++)
                if ((ps = getpwnam(_users[i])) && (ps->pw_uid || _showRoot))
                    insertUser(default_icon, QFile::decodeName(_users[i]), ps);
        }
    }
    endpwent();
    endgrent();
    if (_sortUsers) {
        if (userView)
            userView->sortItems();
        if (userList)
            userList->sort();
    }

    if (!getuid()) {
        seteuid(0);
        setegid(0);
    }
}

void
KGreeter::removeUsers()
{
    if (userView) {
	userView->model()->removeRows(0, userView->model()->rowCount());
    }
}


void
KGreeter::putSession(const QString &type, const QString &name, bool hid, const char *exe)
{
    int prio = exe ? (!strcmp(exe, "default") ? 0 :
                      !strcmp(exe, "custom") ? 1 :
                      !strcmp(exe, "failsafe") ? 3 : 2) : 2;
    for (int i = 0; i < sessionTypes.size(); i++)
        if (sessionTypes[i].type == type) {
            sessionTypes[i].prio = prio;
            return;
        }
    sessionTypes.append(SessType(name, type, hid, prio));
}

void
KGreeter::insertSessions()
{
    for (char **dit = _sessionsDirs; *dit; ++dit)
        foreach (QString ent, QDir(*dit).entryList())
            if (ent.endsWith(".desktop")) {
                KConfigGroup dsk(
                    KSharedConfig::openConfig(
                        QString(*dit).append('/').append(ent)),
                    "Desktop Entry");
                putSession(ent.left(ent.length() - 8),
                            dsk.readEntry("Name"),
                            (dsk.readEntry("Hidden", false) ||
                             (dsk.hasKey("TryExec") &&
                              KStandardDirs::findExe(
                                  dsk.readEntry("TryExec")).isEmpty())),
                            dsk.readEntry("Exec").toLatin1());
            }
    putSession("default", i18nc("@item:inlistbox session type", "Default"), false, "default");
    putSession("custom", i18nc("@item:inlistbox session type", "Custom"), false, "custom");
    putSession("failsafe", i18nc("@item:inlistbox session type", "Failsafe"), false, "failsafe");
    qSort(sessionTypes);
    for (int i = 0; i < sessionTypes.size() && !sessionTypes[i].hid; i++) {
        sessionTypes[i].action = sessGroup->addAction(sessionTypes[i].name);
        sessionTypes[i].action->setData(i);
        sessionTypes[i].action->setCheckable(true);
        switch (sessionTypes[i].prio) {
        case 0: case 1: nSpecials++; break;
        case 2: nNormals++; break;
        }
    }
    sessMenu->addActions(sessGroup->actions());
}

void
KGreeter::slotUserEntered()
{
    struct passwd *pw;

    if (userView) {
        if ((pw = getpwnam(curUser.toLocal8Bit().data()))) {
            QString theUser = QString::fromLocal8Bit(pw->pw_name);
            for (int i = 0, rc = userView->model()->rowCount(); i < rc; i++) {
                UserListViewItem *item =
                    static_cast<UserListViewItem *>(userView->item(i));
                if (item->login == theUser) {
                    userView->setCurrentItem(item);
                    goto oke;
                }
            }
        }
        userView->clearSelection();
    }
  oke:
    if (isVisible())
        slotLoadPrevWM();
    else
        QTimer::singleShot(0, this, SLOT(slotLoadPrevWM()));
}

void
KGreeter::slotUserClicked(QListWidgetItem *item)
{
    if (item) {
        curUser = ((UserListViewItem *)item)->login;
        verify->setUser(curUser);
        slotLoadPrevWM();
    }
}

void
KGreeter::slotSessionSelected()
{
    verify->gplugChanged();
}

void
KGreeter::reject()
{
    verify->reject();
}

void
KGreeter::accept()
{
    verify->accept();
}

void // private
KGreeter::setPrevWM(QAction *wm)
{
    if (curPrev != wm) {
        if (curPrev)
            curPrev->setText(sessionTypes[curPrev->data().toInt()].name);
        if (wm)
            wm->setText(i18nc("@item:inmenu session type",
                              "%1 (previous)",
                              sessionTypes[wm->data().toInt()].name));
        curPrev = wm;
    }
}

void
KGreeter::slotLoadPrevWM()
{
    int len, i, b;
    unsigned long crc, by;
    QByteArray name;
    char *sess;

    // XXX this should actually check for !CoreBusy - would it be safe?
    if (verify->coreState != KGVerify::CoreIdle) {
        needLoad = true;
        return;
    }
    needLoad = false;

    prevValid = true;
    name = curUser.toLocal8Bit();
    gSendInt(G_ReadDmrc);
    gSendStr(name.data());
    gRecvInt(); // ignore status code ...
    if ((len = name.length())) {
        gSendInt(G_GetDmrc);
        gSendStr("Session");
        sess = gRecvStr();
        if (!sess) { /* no such user */
            if (!userView && !userList) { // don't fake if user list shown
                prevValid = false;
                /* simple crc32 */
                for (crc = _forgingSeed, i = 0; i < len; i++) {
                    by = (crc & 255) ^ name[i];
                    for (b = 0; b < 8; b++)
                        by = (by >> 1) ^ (-(by & 1) & 0xedb88320);
                    crc = (crc >> 8) ^ by;
                }
                /* forge a session with this hash - default & custom more probable */
                /* XXX - this should do a statistical analysis of the real users */
#if 1
                setPrevWM(sessionTypes[crc % (nSpecials * 2 + nNormals) % (nSpecials + nNormals)].action);
#else
                i = crc % (nSpecials * 2 + nNormals);
                if (i < nNormals)
                    setPrevWM(sessionTypes[i + nSpecials].action);
                else
                    setPrevWM(sessionTypes[(i - nNormals) / 2].action);
#endif
                return;
            }
        } else {
            for (int i = 0; i < sessionTypes.count() && !sessionTypes[i].hid; i++)
                if (sessionTypes[i].type == sess) {
                    free(sess);
                    setPrevWM(sessionTypes[i].action);
                    return;
                }
            if (!sessGroup->checkedAction())
                KFMsgBox::box(this, sorrybox,
                              i18n("Your saved session type '%1' is not valid any more.\n"
                                   "Please select a new one, otherwise 'default' will be used.", sess));
            free(sess);
            prevValid = false;
        }
    }
    setPrevWM(0);
}

void // protected
KGreeter::pluginSetup()
{
    int field = 0;
    QString ent, pn(verify->pluginName()), dn(dName + '_' + pn);

    if (_preselUser != PRESEL_PREV)
        stsGroup->deleteEntry(verify->entitiesLocal() ? dName : dn, false);
    if (_preselUser != PRESEL_NONE && verify->entityPresettable()) {
        if (verify->entitiesLocal())
            ent = _preselUser == PRESEL_PREV ?
                stsGroup->readEntry(dName, QString()) : _defaultUser;
        else
            ent = _preselUser == PRESEL_PREV ?
                stsGroup->readEntry(dn, QString()) :
                verify->getConf(0, (pn + ".DefaultEntity").toLatin1(), QVariant()).toString();
        field = verify->entitiesFielded() ?
            verify->getConf(0, (pn + ".FocusField").toLatin1(), QVariant(0)).toInt() :
            _focusPasswd;
    }
    verify->presetEntity(ent, field);
    if (userList)
        verify->loadUsers(*userList);
}

void
KGreeter::verifyPluginChanged(int id)
{
    curPlugin = id;
    pluginSetup();
}

void
KGreeter::verifyClear()
{
    curUser.clear();
    slotUserEntered();
    if (QAction *curSel = sessGroup->checkedAction())
        curSel->setChecked(false);
}

void
KGreeter::verifyOk()
{
    if (_preselUser == PRESEL_PREV && verify->entityPresettable())
        stsGroup->writeEntry(verify->entitiesLocal() ?
                                 dName :
                                 dName + '_' + verify->pluginName(),
                             verify->getEntity());
    if (QAction *curSel = sessGroup->checkedAction()) {
        gSendInt(G_PutDmrc);
        gSendStr("Session");
        gSendStr(sessionTypes[curSel->data().toInt()].type.toUtf8());
    } else if (!prevValid) {
        gSendInt(G_PutDmrc);
        gSendStr("Session");
        gSendStr("default");
    }
    done(ex_login);
}

void
KGreeter::verifyFailed()
{
    if (userView)
        userView->setEnabled(false);
    if (needLoad)
        slotLoadPrevWM();
}

void
KGreeter::verifyRetry()
{
    if (userView)
        userView->setEnabled(true);
}

void
KGreeter::verifySetUser(const QString &user)
{
    curUser = user;
    slotUserEntered();
}


KStdGreeter::KStdGreeter()
    : KGreeter()
    , clock(0)
    , pixLabel(0)
{
    QBoxLayout *main_box;
#ifdef WITH_KDM_XCONSOLE
    if (consoleView) {
        QBoxLayout *ex_box = new QVBoxLayout(this);
        main_box = new QHBoxLayout();
        ex_box->addLayout(main_box);
        ex_box->addWidget(consoleView);
    } else
#endif
    {
        main_box = new QHBoxLayout(this);
    }
    int rs = layout()->spacing();
    main_box->setSpacing(layout()->margin());

    if (userView)
        main_box->addWidget(userView);

    QBoxLayout *inner_box = new QVBoxLayout();
    main_box->addLayout(inner_box);
    inner_box->setSpacing(rs);

    if (!_authorized && _authComplain) {
        QLabel *complainLabel = new QLabel(
            i18n("Warning: this is an unsecured session"), this);
        complainLabel->setToolTip(
            i18n("This display requires no X authorization.\n"
                 "This means that anybody can connect to it,\n"
                 "open windows on it or intercept your input."));
        complainLabel->setAlignment(Qt::AlignCenter);
        complainLabel->setFont(*_failFont);
        QPalette p;
        p.setBrush(QPalette::WindowText,
            KColorScheme(QPalette::Active, KColorScheme::Window)
                .foreground(KColorScheme::NegativeText));
        complainLabel->setPalette(p);
        inner_box->addWidget(complainLabel);
    }
    if (!_greetString.isEmpty()) {
        QLabel *welcomeLabel = new QLabel(_greetString, this);
        welcomeLabel->setAlignment(Qt::AlignCenter);
        welcomeLabel->setFont(*_greetFont);
        inner_box->addWidget(welcomeLabel);
    }

    switch (_logoArea) {
    case LOGO_CLOCK:
        clock = new KdmClock(this);
        break;
    case LOGO_LOGO: {
        QMovie *movie = new QMovie(this);
        movie->setFileName(_logo);
        if (movie->isValid()) {
            movie->start();
            pixLabel = new QLabel(this);
            pixLabel->setMovie(movie);
            if (!movie->currentImage().hasAlphaChannel())
                pixLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);
            pixLabel->setIndent(0);
        } else {
            delete movie;
        }
        break; }
    }

    if (userView) {
        if (clock)
            inner_box->addWidget(clock, 0, Qt::AlignCenter);
        else if (pixLabel)
            inner_box->addWidget(pixLabel, 0, Qt::AlignCenter);
        inner_box->addSpacing(inner_box->spacing());
    } else {
        if (clock)
            main_box->addWidget(clock, 0, Qt::AlignCenter);
        else if (pixLabel)
            main_box->addWidget(pixLabel, 0, Qt::AlignCenter);
    }

    goButton = new QPushButton(i18nc("@action:button", "L&ogin"), this);
    goButton->setDefault(true);
    connect(goButton, SIGNAL(clicked()), SLOT(accept()));
    QPushButton *menuButton = new QPushButton(i18nc("@action:button", "&Menu"), this);
    //helpButton

    QWidget *prec;
    if (userView)
        prec = userView;
#ifdef WITH_KDM_XCONSOLE
    else if (consoleView)
        prec = consoleView;
#endif
    else
        prec = menuButton;
    KGStdVerify *sverify =
        new KGStdVerify(this, this, prec, QString(),
                        pluginList, KGreeterPlugin::Authenticate,
                        KGreeterPlugin::Login);
    inner_box->addLayout(sverify->getLayout());
    QMenu *plugMenu = sverify->getPlugMenu();
    sverify->selectPlugin(curPlugin);
    verify = sverify;

    inner_box->addWidget(new KSeparator(Qt::Horizontal, this));

    QBoxLayout *hbox2 = new QHBoxLayout();
    inner_box->addLayout(hbox2);
    hbox2->addWidget(goButton);
    hbox2->addStretch(1);
    hbox2->addWidget(menuButton);
    hbox2->addStretch(1);

    if (sessMenu->actions().count() > 1) {
        inserten(i18nc("@title:menu", "Session &Type"), Qt::ALT + Qt::Key_T, sessMenu);
        needSep = true;
    }

    if (plugMenu) {
        inserten(i18nc("@title:menu", "&Authentication Method"), Qt::ALT + Qt::Key_A, plugMenu);
        needSep = true;
    }

#ifdef XDMCP
    completeMenu(LOGIN_LOCAL_ONLY, ex_choose, i18nc("@action:inmenu", "&Remote Login"), Qt::ALT + Qt::Key_R);
#else
    completeMenu();
#endif

    if (optMenu)
        menuButton->setMenu(optMenu);
    else
        menuButton->hide();

    pluginSetup();

    verify->start();
}

void
KStdGreeter::pluginSetup()
{
    inherited::pluginSetup();
    if (userView) {
        if (verify->entitiesLocal() && verify->entityPresettable())
            userView->show();
        else
            userView->hide();
    }
    adjustGeometry();
    update();
}

void
KStdGreeter::verifyFailed()
{
    goButton->setEnabled(false);
    inherited::verifyFailed();
}

void
KStdGreeter::verifyRetry()
{
    goButton->setEnabled(true);
    inherited::verifyRetry();
}


KThemedGreeter::KThemedGreeter(KdmThemer *_themer)
    : KGreeter(true)
    , themer(_themer)
    , userSliderView(0)
//    , xEventNotifier()
//    , clock(0)
{
    // We do all painting ourselves
    setAttribute(Qt::WA_NoSystemBackground, true);
    // Allow tracking the mouse position
    setMouseTracking(true);

    adjustGeometry();

    themer->setWidget(this);

    if (_allowThemeDebug)
        new QShortcut(QKeySequence(QLatin1String("Ctrl+Alt+D")),
                      this, SLOT(slotDebugToggled()));

    connect(themer, SIGNAL(activated(const QString &)),
            SLOT(slotThemeActivated(const QString &)));

    KdmItem *console_node = themer->findNode("xconsole"); // kdm ext
    KdmItem *console_rect = themer->findNode("xconsole-rect"); // kdm ext
    if (!console_rect)
        console_rect = console_node;
    userlist_node = themer->findNode("userlist");
    userlist_rect = themer->findNode("userlist-rect");
    if (!userlist_rect)
        userlist_rect = userlist_node;
    userslider_node = themer->findNode("userslider");
    userslider_rect = themer->findNode("userslider-rect");
    if (!userslider_rect)
        userslider_rect = userslider_node;

    caps_warning = themer->findNode("caps-lock-warning");
    xauth_warning = themer->findNode("xauth-warning"); // kdm ext
    pam_error = themer->findNode("pam-error");
    KdmLabel *pam_error_label = qobject_cast<KdmLabel *>(pam_error);
    if (pam_error_label)
        pam_error_label->setText(i18n("Login failed"));
    timed_label = themer->findNode("timed-label");

    KdmItem *itm;
    if ((itm = themer->findNode("pam-message"))) // done via msgboxes
        itm->setVisible(false);
    if ((itm = themer->findNode("language_button"))) // not implemented yet
        itm->setVisible(false);

    if (console_node) {
#ifdef WITH_KDM_XCONSOLE
        if (consoleView)
            console_node->setWidget(consoleView);
        else
#endif
            console_rect->setVisible(false);
    }

    if (userslider_node) {
	if (userlist_node)
	    removeUsers();

	userItemDelegate = new UserItemDelegate();
	connect(userItemDelegate, SIGNAL(passwordEntered(const QString &)),
		this, SLOT(slotPasswordEntered(const QString &)));
	connect(userItemDelegate, SIGNAL(tabPressed(const int)),
		this, SLOT(slotTabPressed(const int)));

	userSliderView = new UserSliderView(this);
	userSliderView->setItemDelegate(userItemDelegate);
	imageSize = SLIDER_IMAGE_SIZE;

        //connect(userSliderView, SIGNAL(itemClicked(QListWidgetItem *)),
        //        SLOT(slotUserClicked(QListWidgetItem *)));
        //connect(userSliderView, SIGNAL(itemDoubleClicked(QListWidgetItem *)),
        //        SLOT(accept()));
        connect(userSliderView, SIGNAL(currentItemChanged(QListWidgetItem *, QListWidgetItem *)), 
    		SLOT(slotCurrentItemChanged(QListWidgetItem *, QListWidgetItem *)));

        insertUsers();

	animGroup = new QParallelAnimationGroup;
	animation1 = new UserSliderViewItemAnimation();
	animation2 = new UserSliderViewItemAnimation();
	connect(animation2, SIGNAL(sizeChanged()), userSliderView, SLOT(layoutItems()));
	connect(animation2, SIGNAL(finished()), this, SLOT(slotAnimationFinished()));

        if (userSliderView->count())
    	    userSliderView->setCurrentItem(userSliderView->item(0));

//	connect(&xEventNotifier, SIGNAL(layoutChanged()), this, SLOT(slotLayoutChanged()));
//	connect(&xEventNotifier, SIGNAL(layoutMapChanged()), this, SLOT(slotLayoutChanged()));
//	xEventNotifier.start();
	slotLayoutChanged();
	updateLockStatus();
	//verify->resume();
    }

    if (xauth_warning && (_authorized || !_authComplain))
        xauth_warning->setVisible(false);

//    if (!_greetString.isEmpty()) {
//    }
//    clock = new KdmClock(this, "clock");

    QWidget *prec;
    if (userSliderView)
        prec = userSliderView;
    else if (userView)
        prec = userView;
#ifdef WITH_KDM_XCONSOLE
    else if (consoleView)
        prec = consoleView;
#endif
    else
        prec = 0;
    KGThemedVerify *tverify =
        new KGThemedVerify(this, themer, this, prec, QString(),
                           pluginList, KGreeterPlugin::Authenticate,
                           KGreeterPlugin::Login);
    QMenu *plugMenu = tverify->getPlugMenu();
    tverify->selectPlugin(curPlugin);
    verify = tverify;

    if ((session_button = themer->findNode("session_button"))) {
        if (sessMenu->actions().count() <= 1) {
            session_button->setVisible(false);
            session_button = 0;
        }
    } else {
        if (sessMenu->actions().count() > 1) {
            inserten(i18nc("@title:menu", "Session &Type"), Qt::ALT + Qt::Key_T, sessMenu);
            needSep = true;
        }
    }

    if (plugMenu) {
        inserten(i18nc("@title:menu", "&Authentication Method"), Qt::ALT + Qt::Key_A, plugMenu);
        needSep = true;
    }

#ifdef XDMCP
    completeMenu(LOGIN_LOCAL_ONLY, ex_choose, i18nc("@action:inmenu", "&Remote Login"), Qt::ALT + Qt::Key_R);
#else
    completeMenu();
#endif

    if ((system_button = themer->findNode("system_button"))) {
        if (optMenu)
            addAction(optMenu->menuAction());
        else
            system_button->setVisible(false);
    }

    pluginSetup();

    verify->start();
}

KThemedGreeter::~KThemedGreeter()
{
    if (userSliderView) {
//	xEventNotifier.stop();
	delete animation2;
	delete animation1;
	delete animGroup;
    }
    themer->setWidget(0);
}

void
KThemedGreeter::slotDebugToggled()
{
    if ((debugLevel ^= DEBUG_THEMING))
        themer->slotNeedPlacement();
}

void
KThemedGreeter::insertUser(const QIcon &default_icon,
                     const QString &username, struct passwd *ps)
{

    if (!userSliderView) {
	inherited::insertUser(default_icon, username, ps);
	return;
    }

    if (userList)
        userList->append(username);
    if (!userView)
        return;

    QIcon icon;
    getImage(default_icon, icon, imageSize, ps);

    UserSliderViewItem *item = new UserSliderViewItem(userSliderView, username, icon, username);
}

bool
KThemedGreeter::event(QEvent *e)
{
    if (themer)
        themer->widgetEvent(e);

    if (e->type() == QEvent::KeyPress || e->type() == QEvent::KeyRelease)
        updateLockStatus();

    return inherited::event(e);
}

void
KThemedGreeter::pluginSetup()
{
    inherited::pluginSetup();

    if (userSliderView && verify->entitiesLocal() && verify->entityPresettable() && userslider_node) {
        userslider_node->setWidget(userSliderView);
        userslider_rect->setVisible(true);
        if (userView)
    	    userView->hide();
    	if (userlist_rect)
    	    userlist_rect->setVisible(false);
    } else {
	if (userView && verify->entitiesLocal() && verify->entityPresettable() && userlist_node) {
    	    userlist_node->setWidget(userView);
    	    userlist_rect->setVisible(true);
	} else {
    	    if (userSliderView)
        	userSliderView->hide();
    	    if (userslider_rect)
        	userslider_rect->setVisible(false);
    	    if (userView)
        	userView->hide();
    	    if (userlist_rect)
        	userlist_rect->setVisible(false);
	}
    }
}

void
KThemedGreeter::verifyFailed()
{
    if (userSliderView)
        userSliderView->setEnabled(false);
    inherited::verifyFailed();
}

void
KThemedGreeter::verifyRetry()
{
    if (userSliderView)
        userSliderView->setEnabled(true);
    inherited::verifyRetry();
}

void
KThemedGreeter::updateStatus(bool fail, bool caps, int timedleft)
{
    if (pam_error)
        pam_error->setVisible(fail);
    if (caps_warning)
        caps_warning->setVisible(caps);
    if (timed_label) {
        if (timedleft) {
            if (timedleft != KdmLabel::timedDelay) {
                KdmLabel::timedDelay = timedleft;
                KdmLabel::timedUser = curUser;
                timed_label->setVisible(true);
                timed_label->update();
            }
        } else {
            KdmLabel::timedDelay = -1;
            timed_label->setVisible(false);
        }
    }
}

void
KThemedGreeter::updateLockStatus()
{
    if (userSliderView) {
	unsigned int lmask;
	Window dummy1, dummy2;
	int dummy3, dummy4, dummy5, dummy6;
	XQueryPointer(QX11Info::display(), DefaultRootWindow(QX11Info::display()),
    	            &dummy1, &dummy2, &dummy3, &dummy4, &dummy5, &dummy6,
                &lmask);
	bool capsLocked = lmask & LockMask;

	//qDebug("KThemedGreeter::updateLockStatus %d", capsLocked);
	for (int i = 0; i < userSliderView->model()->rowCount(); i++) {
	    PasswordEditor *editor = static_cast<PasswordEditor*>(userSliderView->itemEditor(i));
	    editor->lineEdit()->setCapsIndicator(capsLocked);
	}
	//userSliderView->layoutItems();
    }
}

void
KThemedGreeter::slotUserEntered()
{

    if (userSliderView) {
	struct passwd *pw;
        if ((pw = getpwnam(curUser.toLocal8Bit().data()))) {
            QString theUser = QString::fromLocal8Bit(pw->pw_name);
            for (int i = 0, rc = userSliderView->model()->rowCount(); i < rc; i++) {
                UserSliderViewItem *item =
                    static_cast<UserSliderViewItem *>(userSliderView->item(i));
                if (item->login == theUser) {
                    userSliderView->setCurrentItem(item);
                    goto oke;
                }
            }
        }
        userSliderView->clearSelection();
	oke:
	    if (isVisible())
    		slotLoadPrevWM();
	    else
    		QTimer::singleShot(0, this, SLOT(slotLoadPrevWM()));
    	return;
    }
    inherited::slotUserEntered();
}

void
KThemedGreeter::slotThemeActivated(const QString &id)
{
    if (id == "login_button")
        accept();
    else if (id == "session_button")
        slotSessMenu();
    else if (id == "system_button")
        slotActionMenu();
    else if (id == "halt_button")
        slotHalt();
    else if (id == "reboot_button")
        slotReboot();
}

void
KThemedGreeter::slotSessMenu()
{
    sessMenu->popup(mapToGlobal(session_button->rect().center()));
}

void
KThemedGreeter::slotActionMenu()
{
    if (system_button)
        optMenu->popup(mapToGlobal(system_button->rect().center()));
    else
        optMenu->popup(mapToGlobal(rect().center()));
}

void
KThemedGreeter::slotHalt()
{
    KDMSlimShutdown::externShutdown(SHUT_HALT, 0, 0, false);
}

void
KThemedGreeter::slotReboot()
{
    KDMSlimShutdown::externShutdown(SHUT_REBOOT, 0, 0, false);
}

void
KThemedGreeter::slotCurrentItemChanged(QListWidgetItem *current, QListWidgetItem *previous)
{
    if (animGroup->indexOfAnimation(animation1) >= 0)
	animGroup->removeAnimation(animation1);
    if (animGroup->indexOfAnimation(animation2) >= 0)
	animGroup->removeAnimation(animation2);
    if (animGroup->state() != QAbstractAnimation::Stopped)
	animGroup->stop();

    int margin  = 3; //style()->pixelMetric(QStyle::PM_FocusFrameHMargin) + 1;
    int editor  = 32;
    int text    = 24;
    int spacing = 2 * margin;
    int frame2  = 4;
    QSize unselectedSize = QSize(imageSize + 16 * margin + frame2,
	imageSize + 16 * margin + frame2 + spacing + text );
    QSize selectedSize = QSize(imageSize * SELECTED_IMAGE_RATIO + 16 * margin + frame2, 
	imageSize * SELECTED_IMAGE_RATIO + spacing + editor + 16 * margin + frame2 + spacing + text);

    if (previous) {
	UserSliderViewItem *prevItem = static_cast<UserSliderViewItem *>(previous);
	animation1->setTargetObject(prevItem);
	animation1->setPropertyName("sizeHint");
	//animation1->setDuration(500);
	animation1->setStartValue(selectedSize);
	animation1->setEndValue(unselectedSize);
	animGroup->addAnimation(animation1);
    }
    if (current) {
	UserSliderViewItem *currItem = static_cast<UserSliderViewItem *>(current);
	animation2->setTargetObject(currItem);
	animation2->setPropertyName("sizeHint");
	//animation2->setDuration(500);
	animation2->setStartValue(unselectedSize);
	animation2->setEndValue(selectedSize);
	animGroup->addAnimation(animation2);
    }
    animGroup->start();
}

void
KThemedGreeter::slotPasswordEntered(const QString &password)
{
    QString user = static_cast<UserSliderViewItem *>(userSliderView->currentItem())->login;
    verify->setUser(user);
    slotLoadPrevWM();
    verify->setPassword(password);
    accept();
}


void
KThemedGreeter::slotTabPressed(const int key)
{
    QListWidgetItem *current = userSliderView->currentItem();
    int next = 0;
    if (current) {
        int count = userSliderView->count();
	int row = userSliderView->row(current);
        switch (key) {
	case Qt::Key_Tab: {
	        next = 0;
		if (row < count - 1)
		    next = row + 1;
    	    }
	    break;
	case Qt::Key_Backtab: {
		next = count - 1;
		if (row > 0)
		    next = row - 1;
	    }
	    break;
	}
    }
    userSliderView->setCurrentItem(userSliderView->item(next));
}

void
KThemedGreeter::slotAnimationFinished()
{
    QListWidgetItem *current = userSliderView->currentItem();
    int row = userSliderView->row(current);
    QWidget *editor = userSliderView->itemEditor(row);
    userSliderView->setFocusProxy(editor);
    editor->setFocus();
}

void
KThemedGreeter::slotLayoutChanged()
{
//    qDebug("KThemedGreeter::slotLayoutChanged");
    if (userSliderView) {
	LayoutUnit layoutUnit = X11Helper::getCurrentLayout();
	QString lang = layoutUnit.layout.left(2);
	for (int i = 0; i < userSliderView->model()->rowCount(); i++) {
	    PasswordEditor *editor = static_cast<PasswordEditor*>(userSliderView->itemEditor(i));
	    editor->lineEdit()->setLangIndicator(lang);
	}
    }
}

void
KThemedGreeter::keyPressEvent(QKeyEvent *e)
{
    inherited::keyPressEvent(e);
    if (!(e->modifiers() & ~Qt::KeypadModifier) &&
        (e->key() == Qt::Key_Return || e->key() == Qt::Key_Enter))
        accept();

    slotLayoutChanged();
}

#include "kgreeter.moc"
