/*

Greeter widget for kdm

Copyright (C) 1997, 1998 Steffen Hansen <hansen@kde.org>
Copyright (C) 2000-2004 Oswald Buddenhagen <ossi@kde.org>
Copyright (C) 2011 ROSA Lab. <ural.mullabaev@rosalab.ru>


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/


#ifndef KGREETER_H
#define KGREETER_H

#include "kgverify.h"
#include "kgdialog.h"

#include <x11_helper.h>

#include <QListWidget>
#include <QListWidgetItem>
#include <QPropertyAnimation>
#include <QStyledItemDelegate>
#include <QLineEdit>

class KdmClock;
class KdmItem;

class KConfigGroup;
class QListWidgetItem;
class QActionGroup;
class QParallelAnimationGroup;
//class XEventNotifier;

class UserSliderViewItem;

class UserListView : public QListWidget {
  public:
    UserListView(QWidget *parent = 0);

    mutable QSize cachedSizeHint;

  protected:
    virtual QSize sizeHint() const;
    virtual void keyPressEvent(QKeyEvent *event);
    virtual void mousePressEvent(QMouseEvent *event);
    virtual void mouseReleaseEvent(QMouseEvent *event);
    virtual void mouseDoubleClickEvent(QMouseEvent *event);

  private:

    bool m_suppressClick;
};

class UserListViewItem : public QListWidgetItem {
  public:
    UserListViewItem(UserListView *parent, const QString &text,
                     const QIcon &icon, const QString &username);

    QString login;
};

class UserSliderView : public UserListView {
    Q_OBJECT
  public:
    UserSliderView(QWidget *parent = 0);

    QWidget *itemEditor(const int row);
    void setItemEditor(const int row, QWidget *widget);

  public Q_SLOTS:
    void layoutItems();

  protected:
    virtual void updateGeometries();
    virtual void resizeEvent(QResizeEvent *event);
    virtual void dragEnterEvent(QDragEnterEvent *event);
    virtual void dragMoveEvent(QDragMoveEvent *event);
    virtual void dragLeaveEvent(QDragLeaveEvent *event);
    virtual void dropEvent(QDropEvent *event);

  private:
    QVector<QWidget *> itemEditors;
};

class UserSliderViewItem : public QObject, public UserListViewItem {
    Q_OBJECT
    Q_PROPERTY(QSize sizeHint READ sizeHint WRITE setSizeHint)
  public:
    UserSliderViewItem(UserSliderView *parent, const QString &text,
                 const QIcon &icon, const QString &username);
    ~UserSliderViewItem();
};

class UserSliderViewItemAnimation : public QPropertyAnimation {
    Q_OBJECT
  public:
    UserSliderViewItemAnimation(QObject *parent = 0);

  Q_SIGNALS:
    void sizeChanged();

  protected:
    void updateCurrentValue(const QVariant &value);
};

class UserItemDelegate : public QStyledItemDelegate
{
    Q_OBJECT

  public:
    UserItemDelegate(QWidget *parent = 0) : QStyledItemDelegate(parent) {}

    void paint(QPainter *painter, const QStyleOptionViewItem &option,
               const QModelIndex &index) const;
    QWidget *createEditor(QWidget *parent, const QStyleOptionViewItem &option,
                          const QModelIndex &index) const;
    void setEditorData(QWidget *editor, const QModelIndex &index) const;
    void setModelData(QWidget *editor, QAbstractItemModel *model,
                      const QModelIndex &index) const;
    void updateEditorGeometry(QWidget * editor, const QStyleOptionViewItem & option,
			      const QModelIndex & index) const;

  Q_SIGNALS:
    void passwordEntered(const QString &password);
    void tabPressed(const int key);

  private slots:
    void slotTextEntered(const QString &password);
    void slotTabPressed(const int key);

  private:
    void drawText(QPainter *p, const QStyleOptionViewItemV4 *option, const QRect &rect) const;
};

class LineEdit : public QLineEdit
{
    Q_OBJECT

  public:
    LineEdit(QWidget *parent = 0);
    ~LineEdit() {};

    void setLangIndicator(const QString &lang);
    void setCapsIndicator(const bool caps);

  Q_SIGNALS:
    void tabPressed(const int key);

  protected:
    virtual bool event(QEvent *event);
    virtual void paintEvent(QPaintEvent *event);

  private:
    QString language;
    bool capsLocked;
};

class PasswordEditor : public QWidget
{
    Q_OBJECT

  public:
    PasswordEditor(QWidget *parent = 0);
    ~PasswordEditor();

    LineEdit* lineEdit() const;

  Q_SIGNALS:
    void textEntered(const QString &text);
    void tabPressed(const int key);

  private slots:
    void slotReturnPressed();
    void slotTabPressed(const int key);

  private:
    QHBoxLayout *layout;
    LineEdit *edit;
    QPushButton *button;
};


struct SessType {
    QString name, type;
    QAction *action;
    bool hid;
    int prio;

    SessType() {}
    SessType(const QString &n, const QString &t, bool h, int p) :
        name(n), type(t), action(0), hid(h), prio(p) {}
    bool operator<(const SessType &st) const {
        return hid != st.hid ? hid < st.hid :
               prio != st.prio ? prio < st.prio :
               name < st.name;
    }
};

class KGreeter : public KGDialog, public KGVerifyHandler {
    Q_OBJECT
    typedef KGDialog inherited;

  public:
    KGreeter(bool themed = false);
    ~KGreeter();

    int imageSize;

  public Q_SLOTS:
    void accept();
    void reject();
    void slotUserClicked(QListWidgetItem *);
    void slotSessionSelected();
    void slotUserEntered();

  protected:
    virtual void insertUser(const QIcon &, const QString &, struct passwd *);
    virtual void insertUsers();
    virtual void removeUsers();
    void putSession(const QString &, const QString &, bool, const char *);
    void insertSessions();
    virtual void pluginSetup();
    void setPrevWM(QAction *);

    QString curUser, dName;
    KConfigGroup *stsGroup;
    UserListView *userView;
    QStringList *userList;
    QMenu *sessMenu;
    QActionGroup *sessGroup;
    QVector<SessType> sessionTypes;
    int nNormals, nSpecials;
    QAction *curPrev;
    bool prevValid;
    bool needLoad;

    static int curPlugin;
    static PluginList pluginList;

  protected Q_SLOTS:
    void slotLoadPrevWM();

  public: // from KGVerifyHandler
    virtual void verifyPluginChanged(int id);
    virtual void verifyClear();
    virtual void verifyOk();
    virtual void verifyFailed();
    virtual void verifyRetry();
    virtual void verifySetUser(const QString &user);
};

class KStdGreeter : public KGreeter {
    Q_OBJECT
    typedef KGreeter inherited;

  public:
    KStdGreeter();

  protected:
    virtual void pluginSetup();

  private:
    KdmClock *clock;
    QLabel *pixLabel;
    QPushButton *goButton;

  public: // from KGVerifyHandler
    virtual void verifyFailed();
    virtual void verifyRetry();
};

class KThemedGreeter : public KGreeter {
    Q_OBJECT
    typedef KGreeter inherited;

  public:
    KThemedGreeter(KdmThemer *_themer);
    ~KThemedGreeter();

  public Q_SLOTS:
    void slotUserEntered();
    void slotThemeActivated(const QString &id);
    void slotSessMenu();
    void slotActionMenu();
    void slotHalt();
    void slotReboot();
    void slotDebugToggled();
    void slotCurrentItemChanged(QListWidgetItem *current, QListWidgetItem *previous);
    void slotPasswordEntered(const QString &password);
    void slotTabPressed(const int key);
    void slotAnimationFinished();
    void slotLayoutChanged();

  protected:
    virtual void insertUser(const QIcon &, const QString &, struct passwd *);
    virtual void updateStatus(bool fail, bool caps, int timedleft);
    virtual void updateLockStatus();
    virtual void pluginSetup();
    virtual void keyPressEvent(QKeyEvent *);
    virtual bool event(QEvent *e);

  private:
//    KdmClock *clock;
    KdmThemer *themer;
    KdmItem *caps_warning, *xauth_warning, *pam_error, *timed_label,
            *userlist_node, *userlist_rect, *userslider_node, *userslider_rect,
            *session_button, *system_button, *halt_button, *reboot_button;

    UserSliderView *userSliderView;
    UserItemDelegate *userItemDelegate;

    QParallelAnimationGroup *animGroup;
    UserSliderViewItemAnimation *animation1;
    UserSliderViewItemAnimation *animation2;

//    XEventNotifier xEventNotifier;

  public: // from KGVerifyHandler
    virtual void verifyFailed();
    virtual void verifyRetry();
};

#endif /* KGREETER_H */
